# This file is part of the project mb-netmgmt
#
# (C) 2022 Deutsche Telekom AG
#
# Deutsche Telekom AG and all other contributors / copyright
# owners license this file to you under the terms of the GPL-2.0:
#
# mb-netmgmt is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as published by
# the Free Software Foundation.
#
# mb-netmgmt is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with mb-netmgmt. If not, see <https://www.gnu.org/licenses/

import importlib
import json
import logging
import sys
import tempfile
import traceback
from socketserver import BaseServer
from urllib.parse import urlparse

import paramiko
import requests


def create_server(protocol, port, callback_url):
    server_address = ("0.0.0.0", port)
    server: BaseServer = protocol.Server(
        server_address, protocol.Handler, bind_and_activate=False
    )
    server.handle_error = handle_error
    server.callback_url = callback_url
    server.allow_reuse_address = True
    server.server_bind()
    server.server_activate()
    return server


def handle_error(request, client_address):
    line = "-" * 40
    logging.error(
        f"""{line}
Exception occurred during processing of request from {client_address}
{traceback.format_exc()}
{line}"""
    )


class Protocol:
    def handle_request(self, request, request_id):
        logging.debug("handle_request: %s", request)
        mb_response = self.post_request(request)
        if "response" not in mb_response:
            self.send_upstream(request, request_id)
        response = self.get_response(mb_response)
        return self.respond(response, request_id)

    def get_response(self, mb_response):
        try:
            return mb_response["response"]
        except KeyError:
            proxy_response = self.read_proxy_response()
            logging.debug("proxy_response: %s", proxy_response)
            return self.post_proxy_response(mb_response, proxy_response)

    def post_proxy_response(self, mb_response, proxy_response):
        response = requests.post(
            mb_response["callbackURL"], json={"proxyResponse": proxy_response}
        )
        response.raise_for_status()
        return response.json()

    def send_upstream(self, request, request_id):
        raise NotImplementedError

    def read_proxy_response(self):
        raise NotImplementedError

    def respond(self, response, request_id):
        raise NotImplementedError

    def post_request(self, request):
        response = requests.post(
            self.callback_url,
            json={"request": request},
        )
        response.raise_for_status()
        return response.json()

    def get_to(self):
        self.key_filename = None
        try:
            imposter_response = requests.get(
                self.server.callback_url.replace("/_requests", "")
            )
            stubs = imposter_response.json()["stubs"]
            proxy = self.get_proxy(stubs[-1])
            if not proxy:
                proxy = self.get_proxy(stubs[0])
            if proxy:
                self.save_key(proxy)
                disable_algorithms(proxy.get("disabled_algorithms", {}))
                return parse_to(proxy["to"])
        except (IndexError, AttributeError):
            pass

    def save_key(self, proxy):
        keyfile = tempfile.NamedTemporaryFile("w")
        try:
            keyfile.write(proxy["key"])
            keyfile.flush()
            self.key_filename = keyfile.name
        except KeyError:
            self.key_filename = None

    def get_proxy(self, stub):
        return stub["responses"][0].get("proxy")


def parse_to(url: str):
    to = urlparse(url)
    if not to.hostname:
        raise ValueError("No hostname")
    return to


def get_cli_patterns():
    patterns = []
    patterns.append(
        b"[\r\n\x1b\[K][\-\w+\.:/]+(?:\([^\)]+\))?[>#] ?$"
    )  # based on IOS driver of Exscript
    patterns.append(
        b"[\r\n\x00\x1b\[K]RP/\d+/(?:RS?P)?\d+\/CPU\d+:[^#]+(?:\([^\)]+\))?#$"
    )  # based on IOS XR driver of Exscript
    patterns.append(
        b'[\r\n\x00\x1b\[K]+(?P<text>[A-Z][\w\/ .:,>\(\)\-\?"]*[^A-Z])(?P<default>\[[\w\/.,():\-]*\])?(?(default)(?P<end1>(?:\?|: ?| |)$)|(?P<end2>: $))'
    )  # Interactive prompt
    patterns.append(b"[\r\n\x00\x1b\[K] --More-- $")  # Terminal paging
    return patterns


def disable_algorithms(disabled_algorithms):
    # https://github.com/ncclient/ncclient/issues/526#issuecomment-1096563028
    class MonkeyPatchedTransport(paramiko.Transport):
        def __init__(self, *args, **kwargs):
            kwargs["disabled_algorithms"] = disabled_algorithms
            super().__init__(*args, **kwargs)

    paramiko.Transport = MonkeyPatchedTransport


if __name__ == "__main__":
    protocol_name = sys.argv[1]
    protocol = importlib.import_module(f"mb_netmgmt.{protocol_name}")

    args = json.loads(sys.argv[2])
    port = args["port"]
    callback_url = args["callbackURLTemplate"].replace(":port", str(port))
    logging.basicConfig(level=args["loglevel"].upper())

    server = create_server(protocol, port, callback_url)
    print(protocol_name, flush=True)
    server.serve_forever()



# This file is part of the project mb-netmgmt
#
# (C) 2022 Deutsche Telekom AG
#
# Deutsche Telekom AG and all other contributors / copyright
# owners license this file to you under the terms of the GPL-2.0:
#
# mb-netmgmt is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as published by
# the Free Software Foundation.
#
# mb-netmgmt is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with mb-netmgmt. If not, see <https://www.gnu.org/licenses/

import re
from socketserver import BaseRequestHandler
from socketserver import ThreadingTCPServer as Server

import paramiko

from mb_netmgmt.__main__ import Protocol, get_cli_patterns

stopped = False


class ParamikoServer(paramiko.ServerInterface):
    def get_allowed_auths(*args):
        return "password,publickey"

    def check_auth_password(*args):
        return paramiko.AUTH_SUCCESSFUL

    def check_auth_publickey(*args):
        return paramiko.AUTH_SUCCESSFUL

    def check_channel_request(*args):
        return paramiko.OPEN_SUCCEEDED

    def check_channel_pty_request(
        self, channel, term, width, height, pixelwidth, pixelheight, modes
    ):
        transport = channel.transport
        channel.upstream = open_upstream(
            transport.to,
            transport.key_filename,
            term,
            width,
            height,
            pixelwidth,
            pixelheight,
        )
        channel.command_prompt = b"#"
        channel.command_prompt = handle_prompt(self.handle_request)
        return True

    def check_channel_shell_request(*args):
        return True

    def check_channel_subsystem_request(*args):
        return True


class Handler(BaseRequestHandler, Protocol):
    def handle(self):
        self.callback_url = self.server.callback_url
        transport = start_server(
            self.request, self.get_to(), self.key_filename, self.handle_request
        )
        self.channel = transport.accept()
        while not stopped:
            request, request_id = self.read_request()
            self.handle_request(request, request_id)

    def send_upstream(self, request, request_id):
        self.channel.upstream.sendall(request["command"])

    def read_request(self):
        request = self.read_message(self.channel, [b"\n", b"\r"])
        return {"command": request.decode()}, None

    def respond(self, response, request_id):
        response = response["response"]
        self.channel.sendall(response)
        return response

    def read_proxy_response(self):
        prompt_patterns = [self.channel.command_prompt]
        prompt_patterns += get_cli_patterns()
        message = self.read_message(self.channel.upstream, prompt_patterns)
        return {"response": message.decode()}

    def read_message(self, channel, patterns):
        message = b""
        end_of_message = False
        while not end_of_message and not stopped:
            message += channel.recv(1024)
            for pattern in patterns:
                if re.findall(pattern, message):
                    end_of_message = True
                    break
        return message


def open_upstream(to, key_filename, term, width, height, pixelwidth, pixelheight):
    if not to:
        return
    client = paramiko.SSHClient()
    client.set_missing_host_key_policy(paramiko.AutoAddPolicy)
    client.connect(
        to.hostname,
        to.port or paramiko.config.SSH_PORT,
        to.username,
        to.password,
        key_filename=key_filename,
        transport_factory=paramiko.Transport,
        look_for_keys=False,
    )
    return client.invoke_shell(term, width, height, pixelwidth, pixelheight)


def handle_prompt(handle_request):
    response = handle_request({"command": ""}, "")
    command_prompt = response.split("\n")[-1].encode()
    return command_prompt


def start_server(request, to, key_filename, handle_request):
    t = paramiko.Transport(request)
    t.add_server_key(paramiko.DSSKey.generate())
    t.add_server_key(paramiko.ECDSAKey.generate())
    t.add_server_key(paramiko.RSAKey.generate(4096))
    t.to = to
    t.key_filename = key_filename
    paramiko_server = ParamikoServer()
    paramiko_server.handle_request = handle_request
    t.start_server(server=paramiko_server)
    return t



# This file is part of the project mb-netmgmt
#
# (C) 2022 Deutsche Telekom AG
#
# Deutsche Telekom AG and all other contributors / copyright
# owners license this file to you under the terms of the GPL-2.0:
#
# mb-netmgmt is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as published by
# the Free Software Foundation.
#
# mb-netmgmt is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with mb-netmgmt. If not, see <https://www.gnu.org/licenses/

"""Network Management Protocols for Mountebank"""
__version__ = "0.0.81"

import os
import subprocess
import time
from contextlib import contextmanager

import requests
from ruamel.yaml import YAML
from ruamel.yaml.scalarstring import (
    DoubleQuotedScalarString,
    LiteralScalarString,
    walk_tree,
)

yaml = YAML()


@contextmanager
def mb(imposters, loglevel="info"):
    with start_mb(loglevel) as mb:
        try:
            put_imposters("localhost", imposters)
            yield mb
        finally:
            mb.terminate()


def start_mb(loglevel="info"):
    mb_netmgmt_dir = os.path.dirname(__file__)
    return subprocess.Popen(
        ["mb", "--loglevel", loglevel, "--pidfile", "/tmp/mb.pid"],
        cwd=mb_netmgmt_dir,
        env=dict(os.environ, PYTHONPATH=mb_netmgmt_dir + "/.."),
    )


def put_imposters(host, imposters, port=2525):
    while True:
        try:
            response = requests.put(
                f"http://{host}:{port}/imposters",
                json={"imposters": imposters},
            )
            break
        except requests.ConnectionError:
            time.sleep(1)
    try:
        response.raise_for_status()
        return response.json()
    except requests.HTTPError as e:
        raise RuntimeError(e.response.json()["errors"])


def dump_imposters(host="localhost", name="imposters", port=2525):
    imposters = get_imposters(host, port)
    use_scalar_strings(imposters)
    yaml.dump(imposters, open(f"{name}.yaml", "w"))


def use_scalar_strings(base):
    walk_tree(
        base,
        {
            "\r": DoubleQuotedScalarString,
            "\n": LiteralScalarString,
        },
    )


def get_imposters(host="localhost", port=2525):
    response = requests.get(
        f"http://{host}:{port}/imposters",
        {"replayable": True, "removeProxies": True},
    )
    return response.json()


def load_imposters(host, name):
    put_imposters(host, read_imposters(name))


def read_imposters(name):
    return yaml.load(open(f"{name}.yaml"))["imposters"]


def proxy_imposters(to, snmp_port=161, telnet_port=23, netconf_port=830):
    return [
        proxy_imposter(snmp_port, "snmp", to, "oids"),
        proxy_imposter(telnet_port, "telnet", f"telnet://{to}", "command"),
        proxy_imposter(
            netconf_port,
            "netconf",
            f"netconf://{os.environ['NETCONF_USERNAME']}:{os.environ['NETCONF_PASSWORD']}@{to}",
            "rpc",
        ),
    ]


def proxy_imposter(port, protocol, to, predicate_generator):
    return {
        "port": port,
        "protocol": protocol,
        "stubs": proxy_stubs(to, predicate_generator),
    }


def proxy_stubs(to, predicate_generator):
    return [
        {
            "responses": [
                {
                    "proxy": {
                        "mode": "proxyAlways",
                        "predicateGenerators": [
                            {"matches": {predicate_generator: True}}
                        ],
                        "to": to,
                    }
                }
            ]
        }
    ]



import io
import os
import re
from threading import Thread
from urllib.parse import urlparse

import ncclient.manager
import paramiko
import pytest
from ncclient.devices.default import DefaultDeviceHandler
from ncclient.transport.session import BASE_NS_1_0, MSG_DELIM, to_ele
from scapy.layers.snmp import ASN1_NULL, SNMPvarbind

from mb_netmgmt import mb, netconf, snmp, ssh, use_scalar_strings, yaml
from mb_netmgmt.__main__ import create_server, get_cli_patterns, parse_to

port = 8081
prompt = b"prompt#"
mock_response = f"""<rpc-reply xmlns="{BASE_NS_1_0}">
  <blubb/>
</rpc-reply>"""
cli_responses = [
    (b"\rIOS-1>", True),
    (b"\rIOS-1#", True),
    (b"\rIOS-1(config)#", True),
    (b"\rIOS-1(config-if)#", True),
    (b"\rRP/0/8/CPU0:IOSXR-2>", True),
    (b"\rRP/0/8/CPU0:IOSXR-2#", True),
    (b"\rRP/0/8/CPU0:IOSXR-2(config)#", True),
    (b"\rRP/0/8/CPU0:IOSXR-2(config-if)#", True),
    (
        b"\rKUncommitted changes found, commit them before exiting(yes/no/cancel)? [cancel]:",
        True,
    ),
    (b"\rProtocol [ipv4]: ", True),
    (b"\rTarget IP address: ", True),
    (b"\rHost name or IP address (control-c to abort): []?", True),
    (b"\rDestination file name (control-c to abort): [running-config]?", True),
    (b"\rDelete net/node0_8_CPU0/disk0:/c12k-mini.vm-4.3.2[confirm]", True),
    (b"\rDestination filename [/net/node0_8_CPU0/disk0:/c12k-mini.vm-4.3.2]?", True),
    (b"\r\n\rCopy : Destination exists, overwrite ?[confirm]", True),
    (b"\rReload hardware module ? [no,yes] ", True),
    (b"\rDo you wish to continue?[confirm(y/n)]", True),
    (b'\r\n\r\nFormat will destroy all data on "harddisk:". Continue? [confirm]', True),
    (b"\r --More-- ", True),
    (b"\rSending 5, 100-byte ICMP Echos to 8.8.8.8, timeout is 2 seconds:", False),
    (b"\rProceed with switchover 0/8/CPU0 -> 0/7/CPU0? [confirm]", True),
    (b"\rPID: SPA-222XXXPOS/RPR, VID: ", False),
    (b"\rNAME: ", False),
    (b"\r  Configured hold time: 180, keepalive: ", False),
    (b"\n\r\n For Address Family: ", False),
]


@pytest.mark.parametrize("protocol", ["http", "snmp", "telnet", "netconf"])
def test_create_imposter(protocol):
    with mb(imposter(protocol, None)):
        pass


def test_ssh():
    with mb(imposter("ssh", [prompt_stub()]), "debug"):
        client = connect_ssh()
        chan = client.invoke_shell()
        out = chan.recv(1024)
        assert out == prompt


def test_ssh_proxy():
    with mb(
        [
            {"protocol": "ssh", "port": 2222, "stubs": [prompt_stub()]},
            {
                "protocol": "ssh",
                "port": port,
                "stubs": [
                    {
                        "responses": [
                            {
                                "proxy": {
                                    "to": f"ssh://{os.environ['NETCONF_USERNAME']}:{os.environ['NETCONF_PASSWORD']}@localhost:2222"
                                }
                            },
                        ]
                    },
                ],
            },
        ]
    ):
        client = connect_ssh()
        chan = client.invoke_shell()
        out = chan.recv(1024)
        assert out == prompt


def prompt_stub():
    return {"responses": [{"is": {"response": prompt.decode()}}]}


def test_create_ssh_server():
    server = create_server(ssh, port, None)
    Thread(target=server.serve_forever).start()
    connect_ssh().close()
    server.shutdown()


def connect_ssh():
    client = paramiko.SSHClient()
    client.set_missing_host_key_policy(paramiko.AutoAddPolicy)
    client.connect("localhost", port)
    return client


def test_create_netconf_server():
    port = 8830
    original_open_upstream = netconf.Handler.open_upstream
    original_post_request = netconf.Handler.post_request
    netconf.Handler.open_upstream = lambda handler: None
    netconf.Handler.post_request = mock_post_request
    server = create_server(netconf, port, None)
    Thread(target=server.serve_forever).start()
    with ncclient.manager.connect(
        host="localhost", port=port, password="", hostkey_verify=False
    ) as m:
        assert m.session_id == "1"
        assert "urn:ietf:params:netconf:base:1.0" in m.server_capabilities
        m.get_config("running")
    netconf.stopped = True
    server.shutdown()
    netconf.stopped = False
    netconf.Handler.open_upstream = original_open_upstream
    netconf.Handler.post_request = original_post_request


def mock_post_request(handler, request):
    if request == {"rpc": ""}:
        return {"response": DefaultDeviceHandler._BASE_CAPABILITIES}
    return {"response": {"rpc-reply": mock_response}}


def test_netconf_upstream():
    from mb_netmgmt.netconf import Handler

    global port
    port = 8831
    Handler.handle = lambda handler: None
    Handler.get_to = lambda handler: urlparse(f"netconf://localhost:{port}")
    handler = Handler(None, None, None)
    handler.save_key({})
    with mb(
        imposter(
            "netconf",
            [
                {
                    "predicates": [{"endsWith": {"rpc": ">running</get-config>"}}],
                    "responses": [
                        {"is": {"rpc-reply": mock_response}},
                    ],
                }
            ],
        ),
        "debug",
    ):
        handler.open_upstream()
        handler.send_upstream(
            {"rpc": "<get-config>running</get-config>"},
            42,
        )
        proxy_response = handler.read_proxy_response()
        assert proxy_response["rpc-reply"] == mock_response


def imposter(protocol, stubs=None):
    return [{"protocol": protocol, "port": port, "stubs": stubs}]


def create_proxy_response(message_id):
    return {
        "response": f'<rpc-reply xmlns="urn:ietf:params:xml:ns:netconf:base:1.0" message-id="{message_id}"/>'
        + MSG_DELIM.decode()
    }


def test_remove_message_id():
    result = netconf.remove_message_id(
        to_ele(f'<rpc-reply xmlns="{BASE_NS_1_0}" message-id="42"><blubb/></rpc-reply>')
    )
    assert result == mock_response


def test_netconf_default_response():
    with mb(imposter("netconf")):
        with ncclient.manager.connect(
            host="localhost", port=port, password="", hostkey_verify=False
        ) as m:
            m.get_config("running")


def test_netconf_private_key():
    with mb(
        [
            {"protocol": "netconf", "port": 8832, "stubs": []},
            {
                "protocol": "netconf",
                "port": port,
                "stubs": [
                    {
                        "responses": [
                            {
                                "proxy": {
                                    "to": f"netconf://{os.environ['NETCONF_USERNAME']}@localhost:8832",
                                    "key": os.environ["NETCONF_KEY"],
                                }
                            },
                        ]
                    },
                ],
            },
        ]
    ):
        ncclient.manager.connect(
            host="localhost", port=port, password="", hostkey_verify=False
        )


@pytest.mark.parametrize(
    ["base", "result"],
    [
        ({"x": "y\nz"}, "x: |-\n  y\n  z\n"),
        ({"x": "y\r\nz"}, 'x: "y\\r\\nz"\n'),
    ],
)
def test_use_scalar_strings(base, result):
    use_scalar_strings(base)

    s = io.StringIO()
    yaml.dump(base, s)
    s.seek(0)
    assert s.read() == result


@pytest.mark.parametrize("cli_response,result", cli_responses)
def test_cli_patterns(cli_response, result):
    matched = False
    for pattern in get_cli_patterns():
        if re.findall(pattern, cli_response):
            matched = True
            break
    assert matched == result


def test_snmp_no_such_instance():
    pkt = b"0\x10\x06\x0c+\x06\x01\x02\x01/\x01\x01\x01\x01\n\x01\x81\x00"
    result = SNMPvarbind(pkt)
    assert not result.noSuchObject
    assert result.noSuchInstance
    assert not result.endOfMibView
    assert bytes(result) == pkt


def test_to_dict():
    varbind = SNMPvarbind()
    result = snmp.to_dict(varbind)
    assert result == {"tag": "NULL", "val": 0}


def test_no_such_instance_to_dict():
    varbind = SNMPvarbind(value=None, noSuchInstance=ASN1_NULL(0))
    result = snmp.to_dict(varbind)
    assert result == {"tag": None, "val": None, "noSuchInstance": 0}


def test_to_varbind():
    result = snmp.to_varbind("1.3", response=dict(val=0, tag="NULL"))
    assert result == SNMPvarbind(oid="1.3", value=0)


def test_to_varbind_no_such_instance():
    result = snmp.to_varbind("1.3", response=dict(val=None, tag=None, noSuchInstance=0))
    assert result == SNMPvarbind(
        oid="1.3", value=None, noSuchObject=None, noSuchInstance=0, endOfMibView=None
    )


def test_parse_to():
    assert parse_to("telnet://localhost").hostname == "localhost"
    assert parse_to("telnet://localhost:23").port == 23
    with pytest.raises(ValueError):
        parse_to("localhost")
